#!/usr/bin/env krubyinit
############################################################################
#    Copyright (C) 2005 by Ian Monroe                                      #
#    ian@monroe.nu                                                         #
#                                                                          #
#    This program is free software; you can redistribute it and#or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################
begin
	require 'Korundum'
rescue LoadError
	error = 'Korundum (KDE bindings for ruby) from kdebindings v3.4 is required for this script.'
	puts error
	`kdialog --sorry '#{error}'`
	exit
end
PLAYLIST = 'playlist.html'
SAVEDIR = Dir.pwd
PLAYLISTNAME = "#{SAVEDIR}/#{PLAYLIST}"
Dir.chdir(File.dirname( File.expand_path(__FILE__) ))
require './xtemplate/lib/xtemplate.rb'
require 'singleton'
require 'amarokwindow.rb'
require 'amarok.rb'
require 'desktopwindow.rb'

#Todo:
#*make a khtml window that works like desktop.py
#*make a non-sucky xtemplate (left as an exercise to the reader)
#*add more to README
#*timestamps
#*a cancel button
#*resize images (use rmagick?)
#*get something to eat

class Uplisty < AmaroKSlots
	def initialize(configWindow, desktopWindow, theCurrentlyPlaying)
		@configWindow = configWindow
		@theCurrentlyPlaying = theCurrentlyPlaying
		@desktopWindow = desktopWindow
	end
	def configure
		##puts "configure"
		@configWindow.show
        @configWindow.setFocus
		#begin
			KDE::Win.activateWindow(@configWindow.winId)
		#rescue NameError
		#	#puts 'No KDE::Win'
		#end
		@configWindow.raise
		#PlaylistConfig.instance.app.updateUserTimestamp(0)
	end
	def stateChange_empty
		##puts "empty"
	end	
	def stateChange_idle
		##puts "idle"
	end
	def stateChange_paused
		##puts "paused"
	end
	def stateChange_playing
		##puts "playing"
	end
	def trackChange
		##puts "track changed"
		@theCurrentlyPlaying.pollAmarok()
		@desktopWindow.updateHtml
	end
	def exit
		PlaylistConfig.instance.app.exit
	end
end

class TrackItem
	attr_reader :artist, :trackTitle, :albumTitle, :duration, :image
	def initialize(artist, trackTitle, albumTitle, duration, image)
		@artist = artist 
		@trackTitle = trackTitle
		@albumTitle = albumTitle
		@duration = duration
		@image = image
	end
	
	def basenameImage
		File.basename(@image)
	end
	
	def basename_hash		
		image = basenameImage()
		image = nil if @image == String.new
		{"artist"=>@artist, "trackTitle"=>@trackTitle, 
		"albumTitle"=>@albumTitle, "duration"=>@duration, 
		"image"=>image}
		#else
		#def to_hash; h = {}; instance_variables.each do |i| h[i] = instance_variable_get i end; return h; end;
	end
	
	def local_hash
		image = @image
		image = nil if @image == String.new
		{"artist"=>@artist, "trackTitle"=>@trackTitle, 
		"albumTitle"=>@albumTitle, "duration"=>@duration, 
		"image"=>image}
	end
end

class GenericPlaylist
	attr_reader :playlistItems
	attr_writer :prevUrl
	attr_reader :prevUrl
	def initialize
		@playlistItems = Array.new
		@uploadMe = Array.new
		@lastImage = String.new
		#@fileData = { :PLAYLIST => @playlistItems, :LASTIMAGE => @lastImage, :PREVURL => @prevUrl }
	end
	def addNewItem(artist, trackTitle, albumTitle, duration, image)
		#I want to make only the latest listing of a track have the cover art
		#if the covers are the same (like when they're playing an album)g
		image = String.new if image.class != String #so, when image is nil
		if image == @lastImage then 
			@lastImage.replace(String.new)
		else
			@uploadMe.unshift(image)
		end
		@lastImage = image
		newItem = TrackItem.new(artist, trackTitle, albumTitle, duration, image)
		@playlistItems.unshift(newItem)
		
		
		#removed old items
		maxItems = PlaylistConfig.instance.maxItems
		##puts "maxItems: #{maxItems.class}"
		@playlistItems.pop() if @playlistItems.size > maxItems
		@uploadMe.pop() if @uploadMe.size > maxItems
	end
	def basename_hash
		a=[]
		playlistItems.each{|x| a.push(x.basename_hash) }
		{ 'tracks' => a }
	end
	
	def local_hash
		a=[]
		playlistItems.each{|x| a.push(x.local_hash) }
		{ 'tracks' => a }
	end
	
	def allImages
		a=[]
		playlistItems.each{|x| a.push(x.image) unless x.image == String.new}
		a
	end	
	
	def imagesKurl
		#a=Array.new
		#playlistItems.each{ |x|
		#	a.push(KDE::URL.new("file:#{x.image}")) unless x.image.class == NilClass
		#}
		#while imageStr = @uploadMe.shift
			#a.unshift(KDE::URL.new("file:#{imageStr}")) 
			#a.unshift(imageStr)
		#end
		ret = Array.new
		ret.replace(@uploadMe)
		@uploadMe.clear
		return ret
	end
		
	def save
		File.open("#{SAVEDIR}/currentplaylist", "w") do |file|
			Marshal.dump({ :PLAYLIST => @playlistItems, :LASTIMAGE => @lastImage, :PREVURL => @prevUrl }, file)
		end
		###puts { :PLAYLIST => @playlistItems, :LASTIMAGE => @lastImage, :PREVURL => @prevUrl }
	end
	
	def load
		begin
			loadedHash = Hash.new
			File.open("#{SAVEDIR}/currentplaylist") do |file|
				loadedHash = Marshal.load(file)
			end
			##puts loadedHash
			#no, I'm not sure why I'm not using Hash.update
			@playlistItems = loadedHash[:PLAYLIST]
			@lastImage = loadedHash[:LASTIMAGE]
			@prevUrl = loadedHash[:PREVURL]
		rescue SystemCallError
			$stderr.print "IO failed: " + $!
		rescue EOFError
			$stderr.print "EOFError: " + $!
		end
	end
end

class HtmlPlaylist < GenericPlaylist

	def initialize
		super
		template = IO.readlines('default.xtemplate')
		#puts template
		@xtemplate = XTemplate::XMLTemplate.new(template)
	end
	def output
		return @xtemplate.expand(basename_hash)
	end
	def outputLocal
		return @xtemplate.expand(local_hash)
	end
end

class SequentialCopy < Qt::Object
	slots 'resultSlot(KIO::Job *)'
	def initialize(fileList, dest)
		super(PlaylistConfig.instance.app)
		@fileList = fileList
		@dest = dest
		@dRPlaylist = KDE::DCOPRef.new("amarok", "playlist")
		uploadNext()
	end
	def uploadNext
		if nextFile = @fileList.pop then
			newJob = KIO::file_copy(nextFile, KDE::URL.new("#{@dest.url}/#{nextFile.fileName}"), -1, true, false,false)
			Qt::Object.connect(newJob, SIGNAL('result(KIO::Job *)'), self, SLOT('resultSlot(KIO::Job *)'))
		end
	end
	def resultSlot(doneJob)
		@dRPlaylist.shortStatusMessage("Uplisting: upload of file resulted in #{doneJob.errorString}") if doneJob.error != 0
		#puts "Uplisting: upload of file resulted in #{doneJob.errorString}"
		uploadNext()
	end
end

class CurrentlyPlaying  < Qt::Object
	slots 'timerEvent()'
	slots "slotEntries(KIO::Job *, const KIO::UDSEntryList &)"
	attr_reader :newDirectory
	attr_writer :newDirectory
	
	def initialize(playlist, configWindow)
		prefs = PlaylistConfig.instance
		super(prefs.app)
		
		@newDirectory = true
		@updateRequired = true
		@prevUrl = String.new
		@playlist = playlist
		@configWindow = configWindow #needed for synchronousRun
		@dR = KDE::DCOPRef.new("amarok", "player")		

		@timer = Qt::Timer.new(self, 'uploader' )
		@timer.start(prefs.updateMin*60*1000)
		prefs.timer = @timer
		prefs.currentlyPlaying = self
		Qt::Object.connect(@timer, SIGNAL("timeout()"), self, SLOT("timerEvent()") )
		
		@directoryContents = Array.new
	end

	def timerEvent
		upload(@playlist.imagesKurl) if @updateRequired
	end

	def getImage()
		image = @dR.coverImage
		if image.index('100@nocover.png')==nil then
			return image
		else			
			return nil
		end
	end
	
	def listImages
		job = KIO::listDir(KDE::URL.new(PlaylistConfig.instance.saveURL))
		Qt::Object.connect(job, SIGNAL("entries(KIO::Job *, const KIO::UDSEntryList &)"),
			self, SLOT("slotEntries(KIO::Job *, const KIO::UDSEntryList &)"))
	end
		
	def slotEntries (job, list)
		@directoryContents.clear
		list.each do |udsentry|
			udsentry.each do |udsatom|				
				@directoryContents.push(udsatom.m_str) if udsatom.m_uds == KIO::UDS_NAME and udsatom.m_str != '.' and udsatom.m_str != '..' and udsatom.m_str != PLAYLIST
			end	
		end
		@newDirectory = false
		upload(@playlist.allImages)
		##puts @playlist.allImages.inspect
 	end
	
	def pollAmarok()
		##puts "polling"
		
		currentUrl = @dR.encodedURL
		##puts "currentURL = #{currentUrl}"
		if(currentUrl != @playlist.prevUrl && currentUrl != nil) then
			@playlist.prevUrl = currentUrl
			@playlist.addNewItem(@dR.artist, 
				@dR.title, 
				@dR.album, 
				@dR.totalTime,
				getImage())
			#upload(@playlist)
			@updateRequired = true
		end
	end
	
	def upload(imageArr)
		unless @newDirectory then
			##puts @directoryContents.inspect
			dest = PlaylistConfig.instance.saveURL
			#playlistName = "#{SAVEDIR}/#{PLAYLIST}"
			File.open(PLAYLISTNAME, 'w') { |f|
				f.puts @playlist.output
			}
			uploadList = Array.new
			uploadList.push(KDE::URL.new(PLAYLISTNAME))
			#KIO::file_copy(KDE::URL.new(PLAYLISTNAME), KDE::URL.new("#{dest}/#{PLAYLISTNAME}"), -1, true, false,false)			
			imageArr.each { |image|
				if not @directoryContents.include?(File.basename(image)) then
					#puts "planning to upload #{image}"
					uploadList.push(KDE::URL.new("file:#{image}"))
					@directoryContents.push(File.basename(image))
				else	
					#puts "not planning to upload #{image}"
				end
			}
			SequentialCopy.new(uploadList, KDE::URL.new("#{dest}"))
			#KIO::copy(uploadList, KDE::URL.new("#{dest}"),false)
			#uploadList.each { |item| 
			#	job = KIO::file_copy(item, KDE::URL.new("#{dest}/#{item.fileName}"), -1, true, false,false)
			#	KIO::NetAccess.synchronousRun(job, @configWindow)
			#	##puts job.errorText
			#	#puts "#{item.url} was copied" 
		#		}
#			UploadThread.new {|ut| ut.run }
			#KIO::CopyJob.new(uploadList, KDE::URL.new("#{dest}"), KIO::CopyJob.Copy, false, false)
			@updateRequired = false
		end
		@playlist.save #might as well save the current state at this point
	end
	

end

class ActualConfig 
	attr_reader :saveURL
	attr_reader :updateMin
	attr_writer :saveURL
	attr_writer :updateMin
	attr_reader :maxItems
	attr_writer :maxItems
    attr_reader :posX
    attr_reader :posY
    attr_reader :sizeX
    attr_reader :sizeY
    attr_writer :posX
    attr_writer :posY
    attr_writer :sizeX
    attr_writer :sizeY
    attr_reader :showDWindow
    attr_writer :showDWindow
	def initialize()
		@saveURL = String.new
		@updateMin = 5
		@maxItems = 50
		@configured = false
        @posX = 100
        @posY = 100
        @sizeX = 400
        @sizeY = 640
        @showDWindow = true
	end
	
	def update(newConfig)
		if newConfig.class == ActualConfig then
			@saveURL = newConfig.saveURL unless newConfig.saveURL == nil
			@updateMin = newConfig.updateMin unless newConfig.updateMin == nil
			@maxItems = newConfig.maxItems unless newConfig.maxItems == nil
            @posX = newConfig.posX unless newConfig.posX == nil
            @posY = newConfig.posY unless newConfig.posY == nil
            @sizeX = newConfig.sizeX unless newConfig.sizeX == nil
            @sizeY = newConfig.sizeY unless newConfig.sizeY == nil
            @showDWindow = newConfig.showDWindow unless newConfig.showDWindow == nil
			@configured = true
		end
	end
end	
class PlaylistConfig
	include Singleton
	

	attr_reader :app
	attr_writer :app #kapplication
	attr_writer :timer #qtimer
    attr_writer :desktopWindow
	attr_writer :currentlyPlaying

	def setConfig(newSaveURL, newUpdateMin, newMaxItems,newPosX,newPosY,newSizeX,newSizeY,newShowDWindow)
		if saveURL() != newSaveURL
			@realConfig.saveURL = newSaveURL
			@currentlyPlaying.listImages
		end
		@realConfig.maxItems = newMaxItems
		if updateMin() != newUpdateMin
			@timer.changeInterval(newUpdateMin*60*1000)
		end
		@realConfig.updateMin = newUpdateMin
        if(@realConfig.posX != newPosX or @realConfig.posY != newPosY or @realConfig.sizeX != newSizeX or @realConfig.sizeY != newSizeY) then
            updateDesktop=true
        end
        if (@realConfig.showDWindow != newShowDWindow)
           @realConfig.showDWindow = newShowDWindow
           if newShowDWindow then @desktopWindow.show
           else @desktopWindow.hide end 
        end
        @realConfig.posX = newPosX
        @realConfig.posY = newPosY
        @realConfig.sizeX = newSizeX
        @realConfig.sizeY = newSizeY
        @desktopWindow.updateSizeNPosition if updateDesktop
	end
    
    
    def posX
        @realConfig.posX
    end
    def showDWindow
        @realConfig.showDWindow
    end
    def posY
        @realConfig.posY
    end
    def sizeX
        @realConfig.sizeX
    end
    def sizeY
        @realConfig.sizeY
    end
	def saveURL
		#init
		@realConfig.saveURL
	end
	
	def updateMin
		@realConfig.updateMin
	end
	
	def maxItems
		@realConfig.maxItems
	end
	
	def playlist
		@realConfig.playlist
	end
	
	def save
		File.open("#{SAVEDIR}/uplistingrc", "w") do |file|
			Marshal.dump(@realConfig, file)
		end
	end
	#asks as a initializer and loads the settings from the disk
	def load()
		#first run stuff that would be the initializer of a normal class
		if (@realConfig.class != ActualConfig) 
			@realConfig = ActualConfig.new()
		end
		#then actually load the file, catching any error since its just a sign that the config
		#hasn't been run before.
		begin
		File.open("#{SAVEDIR}/uplistingrc") do |file|
			@realConfig.update(Marshal.load(file))
		end

		rescue SystemCallError
			$stderr.print "IO failed: " + $!
		else
			##puts "saveURL set as #{saveURL}"
		end
	end

end

class MainApp	

	def initialize
		#setup app
		about = KDE::AboutData.new("uplisting", "Uplisting", "0.1")
		KDE::CmdLineArgs.init(ARGV,about)
		@app = KDE::Application.new
		#playlistconfig needs to be setup before AmarokWindow since the latter loads the config from the former
		theConfig = PlaylistConfig.instance
		theConfig.app = @app
		theConfig.load()
		#now that we've loaded the settings, setup the playlist
		#setup the main window
		playlist = HtmlPlaylist.new
		playlist.load()
        root = Qt::Widget.new(nil)
        
		desktopWin = DesktopWindow.new(nil,playlist)
        configWindow = AmarokWindow.new(desktopWin,'amarokwindow',true,Qt::WType_Dialog  )
        theConfig.desktopWindow = desktopWin
		
		theCurrentlyPlaying =  CurrentlyPlaying.new(playlist,configWindow)
		theCurrentlyPlaying.pollAmarok()
		desktopWin.updateHtml()
		theCurrentlyPlaying.listImages()
		desktopWin.show if theConfig.showDWindow
		#desktopWin.lower
		#if the program doesn't have a location to save
		if theConfig.saveURL == String.new then
			configWindow.show
		end		
		amaroKInterface = AmaroKLib.new(@app, Uplisty.new(configWindow, desktopWin, theCurrentlyPlaying))
		@app.exec
	end
end
m = MainApp.new
